user_list: Make list head static.
[paraslash.git] / ipc.c
diff --git a/ipc.c b/ipc.c
index 37d30715adaf02bf4e0926fcedfc1c8aa4be1949..85013a636cb3a7f634d190d9044808454274efd8 100644 (file)
--- a/ipc.c
+++ b/ipc.c
@@ -1,14 +1,14 @@
-/*
- * Copyright (C) 2006-2008 Andre Noll <maan@systemlinux.org>
- *
- * Licensed under the GPL v2. For licencing details see COPYING.
- */
+/* Copyright (C) 2006 Andre Noll <maan@tuebingen.mpg.de>, see file COPYING. */
 
 /** \file ipc.c Inter-process communication and shared memory helpers. */
 
 #include "para.h"
 #include "error.h"
 #include "ipc.h"
+#include <sys/types.h>
+#include <sys/param.h>
+#include <sys/sysctl.h>
+
 #include <sys/ipc.h>
 #include <sys/shm.h>
 #include <sys/sem.h>
@@ -16,7 +16,7 @@
 /**
  * Define a new mutex.
  *
* \return The identifier for the new mutex on success, \a -E_SEM_GET
  * on errors.
  *
  * \sa semget(2).
@@ -24,7 +24,7 @@
 int mutex_new(void)
 {
        int ret = semget(IPC_PRIVATE, 1, IPC_CREAT | 0666);
-       return ret < 0?  -E_SEM_GET : ret;
+       return ret < 0? -ERRNO_TO_PARA_ERROR(errno) : ret;
 }
 
 /**
@@ -32,14 +32,14 @@ int mutex_new(void)
  *
  * \param id The identifier of the mutex to be destroyed.
  *
- * \return Positive on success, \a -E_SEM_REMOVE on errors.
+ * \return Standard.
  *
  * \sa semctl(2)
  */
 int mutex_destroy(int id)
 {
        int ret = semctl(id, 0, IPC_RMID);
-       return ret < 0? -E_SEM_REMOVE : 1;
+       return ret < 0? -ERRNO_TO_PARA_ERROR(errno) : 1;
 }
 
 static void para_semop(int id, struct sembuf *sops, int num)
@@ -49,7 +49,7 @@ static void para_semop(int id, struct sembuf *sops, int num)
                        return;
        } while (errno == EINTR);
        if (errno == EIDRM) {
-               PARA_NOTICE_LOG("semaphore set %d was removed\n", id);
+               PARA_CRIT_LOG("semaphore set %d was removed\n", id);
                return;
        }
        PARA_EMERG_LOG("fatal semop error %s: pid %d\n", strerror(errno),
@@ -64,7 +64,7 @@ static void para_semop(int id, struct sembuf *sops, int num)
  *
  * This function either succeeds or aborts.
  *
- * \sa semop(2), struct misc_meta_data.
+ * \sa semop(2), struct \ref misc_meta_data.
  */
 void mutex_lock(int id)
 {
@@ -90,7 +90,7 @@ void mutex_lock(int id)
  *
  * This function either succeeds or aborts.
  *
- * \sa semop(2), struct misc_meta_data.
+ * \sa semop(2), struct \ref misc_meta_data.
  */
 void mutex_unlock(int id)
 {
@@ -109,14 +109,15 @@ void mutex_unlock(int id)
  *
  * \param size The size of the shared memory area to create.
  *
* \return The id of the shared memory areay on success, \a -E_SHM_GET on errors.
  *
  * \sa shmget(2).
  */
 int shm_new(size_t size)
 {
        int ret = shmget(IPC_PRIVATE, size, IPC_CREAT | IPC_EXCL | 0600);
-       return ret < 0 ? -E_SHM_GET : ret;
+       return ret < 0 ? -ERRNO_TO_PARA_ERROR(errno) : ret;
 }
 
 /**
@@ -125,7 +126,7 @@ int shm_new(size_t size)
  * \param id The shared memory identifier.
  *
  * \return The return value of the underlying shmctl() call on success,
- * \a -E_SHM_DESTROY on errors.
+ * a negative error code on errors.
  *
  * \sa shmctl(2).
  */
@@ -133,28 +134,48 @@ int shm_destroy(int id)
 {
        struct shmid_ds shm_desc;
        int ret = shmctl(id, IPC_RMID, &shm_desc);
-       return ret < 0? -E_SHM_DESTROY : ret;
+       return ret < 0? -ERRNO_TO_PARA_ERROR(errno) : ret;
 }
 
 /**
  * Attach a shared memory segment.
  *
  * \param id The identifier of the shared memory segment to attach.
- * \param mode either ATTACH_RO (read only) or ATTACH_RW (read/write).
- * \param result points to the attached area just attached.
+ * \param mode Either ATTACH_RO (read only) or ATTACH_RW (read/write).
+ * \param result Points to the attached area just attached on success.
  *
- * \return positive on success, \a -E_SHM_ATTACH on errors.
+ * \return Standard.
  *
  * \sa shmat(2).
  */
 int shm_attach(int id, enum shm_attach_mode mode, void **result)
 {
-       if (mode == ATTACH_RW) {
+       if (mode == ATTACH_RW)
                *result = shmat(id, NULL, 0);
-               return *result? 1 : -E_SHM_ATTACH;
-       }
-       *result = shmat(id, NULL, SHM_RDONLY);
-       return *result? 1 : -E_SHM_ATTACH;
+       else
+               *result = shmat(id, NULL, SHM_RDONLY);
+       return *result == (void *) -1? -ERRNO_TO_PARA_ERROR(errno) : 1;
+}
+
+/**
+ * Get the size of a shared memory segment.
+ *
+ * \param id The shared memory segment identifier.
+ * \param result Size in bytes is returned here, zero on errors.
+ *
+ * \return Standard.
+ *
+ * \sa shmctl(2).
+ */
+int shm_size(int id, size_t *result)
+{
+       struct shmid_ds ds; /* data structure */
+
+       *result = 0;
+       if (shmctl(id, IPC_STAT, &ds) < 0)
+               return -ERRNO_TO_PARA_ERROR(errno);
+       *result = ds.shm_segsz;
+       return 1;
 }
 
 /**
@@ -162,12 +183,64 @@ int shm_attach(int id, enum shm_attach_mode mode, void **result)
  *
  * \param addr The address of the attached segment.
  *
- * \return Positive on success, \a -E_SHM_DETACH on errors.
+ * \return Standard.
  *
  * \sa shmdt(2).
  */
 int shm_detach(void *addr)
 {
        int ret = shmdt(addr);
-       return ret < 0? -E_SHM_DETACH : 1;
+       return ret < 0? -ERRNO_TO_PARA_ERROR(errno) : 1;
+}
+
+# if defined __FreeBSD__ || defined __NetBSD__
+#      define SYSCTL_SHMMAX_VARIABLE "kern.ipc.shmmax"
+# elif defined __APPLE__
+#      define SYSCTL_SHMMAX_VARIABLE "kern.sysv.shmmax"
+# else
+#      undef SYSCTL_SHMMAX_VARIABLE
+# endif
+
+/**
+ * Get the maximal size of a shared memory area.
+ *
+ * The value is only computed once when the function is called for the first
+ * time.  Subsequent calls return the number which was computed during the
+ * first call.
+ *
+ * \return A number suitable as an argument to \ref shm_new().
+ */
+size_t shm_get_shmmax(void)
+{
+       static size_t shmmax;
+
+       if (shmmax > 0) /* only dance once */
+               return shmmax;
+#ifdef __linux__ /* get it from proc fs */
+       {
+               int fd = open("/proc/sys/kernel/shmmax", O_RDONLY);
+               if (fd >= 0) {
+                       char buf[100] = "";
+                       int ret = read(fd, buf, sizeof(buf) - 1);
+                       if (ret > 0) {
+                               buf[ret] = '\0';
+                               shmmax = strtoul(buf, NULL, 10);
+                       }
+                       close(fd);
+               }
+       }
+#elif defined SYSCTL_SHMMAX_VARIABLE
+       {
+               size_t len = sizeof(shmmax);
+               sysctlbyname(SYSCTL_SHMMAX_VARIABLE, &shmmax, &len, NULL, 0);
+       }
+#elif defined SHMMAX
+       shmmax = SHMMAX;
+#endif
+       if (shmmax == 0) {
+               PARA_WARNING_LOG("unable to determine shmmax\n");
+               shmmax = 65535; /* last resort */
+       }
+       PARA_INFO_LOG("shmmax: %zu\n", shmmax);
+       return shmmax;
 }
@@ -16,7 +16,7 @@
  * Define a new mutex.
  *
- * \return The identifier for the new mutex on success, \a -E_SEM_GET
+ * \return The identifier for the new mutex on success, a negative error code
  * on errors.
  *
@@ -109,14 +109,15 @@ void mutex_unlock(int id)
  * \param size The size of the shared memory area to create.
  *
- * \return The id of the shared memory areay on success, \a -E_SHM_GET on errors.
+ * \return The id of the shared memory array on success, a negative error
+ * code on errors.
  *
  * \sa shmget(2).